VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 1  'Persistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "DirectoryInfo"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
'    CopyRight (c) 2005 Kelly Ethridge
'
'    This file is part of VBCorLib.
'
'    VBCorLib is free software; you can redistribute it and/or modify
'    it under the terms of the GNU Library General Public License as published by
'    the Free Software Foundation; either version 2.1 of the License, or
'    (at your option) any later version.
'
'    VBCorLib is distributed in the hope that it will be useful,
'    but WITHOUT ANY WARRANTY; without even the implied warranty of
'    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'    GNU Library General Public License for more details.
'
'    You should have received a copy of the GNU Library General Public License
'    along with Foobar; if not, write to the Free Software
'    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
'
'    Module: DirectoryInfo
'

''
' Represents a directory to modify and retrieve information about.
'
' @remarks The DirectoryInfo object does not verify the existence
' of the directory until information has to be read or written to that directory.
'
' @see Constructors
'
Option Explicit
Implements IObject
Implements FileSystemInfo

Private mPath           As String
Private mOriginalPath   As String
Private mIsDirty        As Boolean
Private mFileAttributes As Long
Private mCreationTime   As cDateTime
Private mLastAccessTime As cDateTime
Private mLastWriteTime  As cDateTime


''
' Returns the name of the directory this instance represents.
'
' @param The name of the directory.
'
Public Property Get Name() As String
    Name = Path.GetFileName(mPath)
    If Len(Name) = 0 Then Name = mPath
End Property

''
' Returns the name of the directory, including the full path to the directory.
'
' @param The full path to the directory.
'
Public Property Get FullName() As String
    FullName = mPath
End Property

''
' Returns the directory's attributes (ReadOnly, Hidden, Archive, ...)
'
' @return The directories attributes since the last time Refresh was called.
' @remarks The attributes are not kept current. If the they are changed since
' last checked, then Refresh must be called to retrieve the current settings.
'
Public Property Get Attributes() As FileAttributes
    If mIsDirty Then Call Refresh
    Attributes = mFileAttributes
End Property

''
' Sets the attributes for the directory.
'
' @param RHS The attributes to set for the directory.
' @remarks Setting just one attribute will clear all the other attributes. To
' preserve any attributes, then OR must be used with the attributes to
' be preserved.
'
Public Property Let Attributes(ByVal RHS As FileAttributes)
    Call File.SetAttributes(mPath, RHS)
    Call MarkDirty
End Property

''
' Returns the creation time for this directory in local time.
'
' @return A cDateTime object representing the creation time for this directory.
' @remarks The time returned is the UTC time with the timezone offset applied.
' For Pacific Coast, a -8 hours is added to the time before returning.
'
Public Property Get CreationTime() As Variant
    If mIsDirty Then Refresh
    Set CreationTime = mCreationTime
End Property

''
' Sets the creation time for the directory in local time.
'
' @param RHS A Date of the new time.
' @remarks The time has the UTC timezone offset applied to it before
' being written to the directory.
'
Public Property Let CreationTime(ByVal RHS As Variant)
    Set Me.CreationTime = cDateTime.GetcDateTime(RHS)
End Property

''
' Sets the creation time for the directory in local time.
'
' @param RHS A cDateTime object of the new time.
' @remarks The time has the UTC timezone offset applied to it before
' being written to the directory.
'
Public Property Set CreationTime(ByVal RHS As Variant)
    If Not TypeOf RHS Is cDateTime Then _
        Throw Cor.NewArgumentException(Environment.GetResourceString(Argument_DateRequired), "CreationTime")

    Call Directory.SetCreationTime(mPath, mCreationTime)
    Call MarkDirty
End Property

''
' Returns the creation time as UTC time.
'
' @return A cDateTime object of the UTC time.
' @remarks The time returned is the time on the directory without
' any offsets being applied.
'
Public Property Get CreationTimeUtc() As Variant
    If mIsDirty Then Refresh
    Set CreationTimeUtc = mCreationTime.ToUniversalTime
End Property

''
' Sets the creation time as UTC time.
'
' @param RHS A Date value as the new time.
' @remarks The time has the UTC timezone offset removed from it before
' being written to the directory.
'
Public Property Let CreationTimeUtc(ByVal RHS As Variant)
    Set Me.CreationTimeUtc = cDateTime.GetcDateTime(RHS)
End Property

''
' Sets the creation time as UTC time.
'
' @param RHS A cDateTime object of the new time.
' @remarks The time has the UTC timezone offset removed from it before
' being written to the directory.
'
Public Property Set CreationTimeUtc(ByVal RHS As Variant)
    If Not TypeOf RHS Is cDateTime Then _
        Throw Cor.NewArgumentException(Environment.GetResourceString(Argument_DateRequired), "CreationTime")

    Call Directory.SetCreationTimeUtc(mPath, RHS)
    Call MarkDirty
End Property

''
' Returns the last time the directory was accessed in local time.
'
' @return A cDateTime object of the time.
' @remarks The returned time is the UTC time with the UTC timezone offset applied.
'
Public Property Get LastAccessTime() As Variant
    If mIsDirty Then Refresh
    Set LastAccessTime = mLastAccessTime
End Property

''
' Sets the time the directory was last accessed in local time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Let LastAccessTime(ByVal RHS As Variant)
    Set Me.LastAccessTime = cDateTime.GetcDateTime(RHS)
End Property

''
' Sets the time the directory was last accessed in local time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Set LastAccessTime(ByVal RHS As Variant)
    If Not TypeOf RHS Is cDateTime Then _
        Throw Cor.NewArgumentException(Environment.GetResourceString(Argument_DateRequired), "LastAccessTime")
        
    Call Directory.SetLastAccessTime(mPath, mLastAccessTime)
    Call MarkDirty
End Property

''
' Returns the last time the directory was accessed in UTC time.
'
' @return a cDateTime object of the time.
'
Public Property Get LastAccessTimeUtc() As Variant
    If mIsDirty Then Refresh
    Set LastAccessTimeUtc = mLastAccessTime.ToUniversalTime
End Property

''
' Sets the last time the directory was accessed in UTC time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Let LastAccessTimeUtc(ByVal RHS As Variant)
    Set Me.LastAccessTimeUtc = cDateTime.GetcDateTime(RHS)
End Property

''
' Sets the last time the directory was accessed in UTC time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Set LastAccessTimeUtc(ByVal RHS As Variant)
    If Not TypeOf RHS Is cDateTime Then _
        Throw Cor.NewArgumentException(Environment.GetResourceString(Argument_DateRequired), "LastAccessTime")
    
    Call Directory.SetLastAccessTimeUtc(mPath, RHS)
    Call MarkDirty
End Property

''
' Returns the last time the directory was written to in local time.
'
' @return A cDateTime object of the time.
'
Public Property Get LastWriteTime() As Variant
    If mIsDirty Then Refresh
    Set LastWriteTime = mLastWriteTime
End Property

''
' Sets the last time the directory was written to in local time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Let LastWriteTime(ByVal RHS As Variant)
    Set Me.LastAccessTime = cDateTime.GetcDateTime(RHS)
End Property

''
' Sets the last time the directory was written to in local time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Set LastWriteTime(ByVal RHS As Variant)
    If Not TypeOf RHS Is cDateTime Then _
        Throw Cor.NewArgumentException(Environment.GetResourceString(Argument_DateRequired), "LastWriteTime")
    
    Call Directory.SetLastWriteTime(mPath, mLastWriteTime)
    Call MarkDirty
End Property

''
' Returns the last time the directory was written to in UTC time.
'
' @return A cDateTime object of the time.
'
Public Property Get LastWriteTimeUtc() As Variant
    If mIsDirty Then Refresh
    Set LastWriteTimeUtc = mLastWriteTime.ToUniversalTime
End Property

''
' Sets the last time the directory was written to in UTC time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Let LastWriteTimeUtc(ByVal RHS As Variant)
    Set Me.LastWriteTimeUtc = cDateTime.GetcDateTime(RHS)
End Property

''
' Sets the last time the directory was written to in UTC time.
'
' @param RHS A Date or cDateTime object of the new time.
'
Public Property Set LastWriteTimeUtc(ByVal RHS As Variant)
    If Not TypeOf RHS Is cDateTime Then _
        Throw Cor.NewArgumentException(Environment.GetResourceString(Argument_DateRequired), "LastAccessTime")
    
    Call Directory.SetLastWriteTimeUtc(mPath, RHS)
    Call MarkDirty
End Property

''
' Reloads all of the attributes and timestamps for the directory.
'
Public Sub Refresh()
    Dim Data As WIN32_FILE_ATTRIBUTE_DATA
    Dim e As Long
    
    e = File.GetFileData(mPath, Data)
    If e <> NO_ERROR Then IOError e, mPath
    
    With Data
        mFileAttributes = .dwFileAttributes
        Set mCreationTime = cDateTime.FromFileTime(.ftCreationTime)
        Set mLastAccessTime = cDateTime.FromFileTime(.ftLastAccessTime)
        Set mLastWriteTime = cDateTime.FromFileTime(.ftLastWriteTime)
    End With
    
    mIsDirty = False
End Sub

''
' Checks if the directory actually exists.
'
' @return An indication of the directory existing.
'
Public Property Get Exists() As Boolean
    Exists = Directory.Exists(mPath)
End Property

''
' Returns an empty string for directories.
'
' @return An empty string.
'
Public Property Get Extension() As String
' do nothing
End Property

''
' Returns the parent directory as a DirectoryInfo object.
'
' @return A DirectroyInfo object of the parent directory, or Nothing if
' the original directory is a root, such as 'c:\'
Public Property Get Parent() As DirectoryInfo
    Dim p As String
    
    p = Path.GetDirectoryName(mPath)
    If Len(p) > 0 Then Set Parent = Cor.NewDirectoryInfo(p)
End Property

''
' Returns a DirectoryInfo object of the root directory of this instance.
'
' @return A DirectoryInfo object of the root.
'
Public Property Get Root() As DirectoryInfo
    Set Root = Cor.NewDirectoryInfo(Path.GetPathRoot(mPath))
End Property

''
' Creates the path represented by this instance.
'
Public Sub Create()
    Call Directory.CreateDirectory(mPath)
End Sub

''
' Creates a directory within this instance, returning the new subdirectory.
'
' @param sPath The new subdirectory to be created.
' @return A DirectoryInfo object of the new subdirectory path.
'
Public Function CreateSubdirectory(ByVal sPath As String) As DirectoryInfo
    Set CreateSubdirectory = Directory.CreateDirectory(Path.Combine(mPath, sPath))
End Function

''
' Deletes this directory if it is empty, or all of its contents and itself
' if Recursive is set to True.
'
' @param Recursive Indicates if all entries in the directory should be deleted.
'
Public Sub Delete(Optional ByVal Recursive As Boolean)
    Call Directory.Delete(mPath, Recursive)
End Sub

''
' Returns a list of directories within this directory.
'
' @param SearchPattern A pattern for all directories to match.
' @return A list of all matching directories.
' @remarks If no matching directories were found, then an empty array is returned.
'
Public Function GetDirectories(Optional ByVal SearchPattern As String = "*") As DirectoryInfo()
    Dim Ret() As DirectoryInfo
    
    Dim Paths() As String
    Paths = Directory.GetDirectories(mPath, SearchPattern)
    
    Dim NumPaths As Long
    NumPaths = cArray.GetLength(Paths)
    If NumPaths > 0 Then
        Dim i As Long
        ReDim Ret(0 To NumPaths - 1)
        For i = 0 To NumPaths - 1
            Set Ret(i) = Cor.NewDirectoryInfo(Path.Combine(mPath, Paths(i)))
        Next i
    Else
        ' No subdirectories were found.
        ' LAME SPEC: .NET actually returns a zero-length array, where the
        ' MSDN says it returns the root directory.
        Ret = cArray.CreateInstance(ciObject)
    End If
    GetDirectories = Ret
End Function

''
' Moves this directory and its contents to the specified path.
'
' @param DestDirName The new path and name of the directory.
'
Public Sub MoveTo(ByVal DestDirName As String)
    Call Directory.Move(mPath, DestDirName)
    mPath = Path.GetFullPath(DestDirName)
    mOriginalPath = Path.GetDirectoryName(mPath)
End Sub

''
' Returns a list of file matching the pattern in this directory.
'
' @param SearchPattern The pattern the files must match.
' @return A list of files that matched the pattern.
' @remarks If no files matched the pattern, a zero-length array is returned.
'
Public Function GetFiles(Optional ByVal SearchPattern As String = "*") As FileInfo()
    Dim Ret() As FileInfo
    
    Dim Files() As String
    Files = Directory.GetFiles(mPath, SearchPattern)
    
    Dim NumFiles As Long
    NumFiles = cArray.GetLength(Files)
    If NumFiles > 0 Then
        Dim i As Long
        ReDim Ret(0 To NumFiles - 1)
        For i = 0 To NumFiles - 1
            Set Ret(i) = Cor.NewFileInfo(Path.Combine(mPath, Files(i)))
        Next i
    Else
        Ret = cArray.CreateInstance(ciObject)
    End If
    GetFiles = Ret
End Function

''
' Returns a list of all entries in the directory that match the pattern.
'
' @param SearchPattern The pattern to match against all entries.
' @return A list of all entries that matched the pattern.
' @remarks If no entries matched the pattern, a zero-length array is returned.
'
Public Function GetFileSystemInfos(Optional ByVal SearchPattern As String = "*") As FileSystemInfo()
    Dim Files() As String
    Files = Directory.GetFiles(mPath, SearchPattern)
    
    Dim Folders() As String
    Folders = Directory.GetDirectories(mPath, SearchPattern)
    
    Dim InfoCount As Long
    InfoCount = cArray.GetLength(Files) + cArray.GetLength(Folders)
    
    Dim Ret() As FileSystemInfo
    If InfoCount > 0 Then
        Dim i As Long
        ReDim Ret(0 To InfoCount - 1)
        For i = 0 To UBound(Files)
            Set Ret(i) = Cor.NewFileInfo(Path.Combine(mPath, Files(i)))
        Next i
        
        Dim j As Long
        j = UBound(Files) + 1
        For i = 0 To UBound(Folders)
            Set Ret(j + i) = Cor.NewDirectoryInfo(Path.Combine(mPath, Folders(i)))
        Next i
    Else
        ' None found, so return an empty array.
        Ret = cArray.CreateInstance(ciObject)
    End If
    
    GetFileSystemInfos = Ret
End Function

''
' Returns a string representation of this object instance.
'
' @return String representing this instance.
Public Function ToString() As String
    ToString = mOriginalPath
End Function

''
' Returns a boolean indicating if the value and this object
' instance are the same instance.
'
' @param value The value to compare equality to.
' @return Boolean indicating equality.
Public Function Equals(ByRef Value As Variant) As Boolean
    Equals = Object.Equals(Me, Value)
End Function

''
' Returns a pseudo-unique number identifying this instance.
'
' @return Pseudo-unique number identifying this instance.
Public Function GetHashCode() As Long
    GetHashCode = ObjPtr(CUnk(Me))
End Function


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'   Friend Interface
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Friend Sub Init(ByVal sPath As String)
    Call Path.VerifyPath(sPath)
    mOriginalPath = sPath
    If Not Path.IsPathRooted(sPath) Then
        mPath = Path.GetFullPath(sPath)
    Else
        mPath = sPath
    End If
End Sub


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'   Private Helpers
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Sub MarkDirty()
    mIsDirty = True
End Sub


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'   Class Events
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Sub Class_Initialize()
    Call MarkDirty
End Sub

Private Sub Class_ReadProperties(PropBag As PropertyBag)
    Call Init(PropBag.ReadProperty("Path"))
End Sub

Private Sub Class_WriteProperties(PropBag As PropertyBag)
    Call PropBag.WriteProperty("Path", mOriginalPath)
End Sub


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'   IObject Interface
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Function IObject_Equals(Value As Variant) As Boolean
    IObject_Equals = Equals(Value)
End Function

Private Function IObject_GetHashcode() As Long
    IObject_GetHashcode = GetHashCode
End Function

Private Function IObject_ToString() As String
    IObject_ToString = ToString
End Function


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'   FileSystemInfo Interface
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
Private Property Get FileSystemInfo_Attributes() As FileAttributes
    FileSystemInfo_Attributes = Attributes
End Property

Private Property Let FileSystemInfo_Attributes(ByVal RHS As FileAttributes)
    Attributes = RHS
End Property

Private Property Set FileSystemInfo_CreationTime(ByVal RHS As Variant)
    Set CreationTime = RHS
End Property

Private Property Let FileSystemInfo_CreationTime(ByVal RHS As Variant)
    CreationTime = RHS
End Property

Private Property Get FileSystemInfo_CreationTime() As Variant
    Set FileSystemInfo_CreationTime = CreationTime
End Property

Private Property Set FileSystemInfo_CreationTimeUtc(ByVal RHS As Variant)
    Set CreationTimeUtc = RHS
End Property

Private Property Let FileSystemInfo_CreationTimeUtc(ByVal RHS As Variant)
    CreationTimeUtc = RHS
End Property

Private Property Get FileSystemInfo_CreationTimeUtc() As Variant
    Set FileSystemInfo_CreationTimeUtc = CreationTimeUtc
End Property

Private Sub FileSystemInfo_Delete()
    Call Delete
End Sub

Private Function FileSystemInfo_Equals(Value As Variant) As Boolean
    FileSystemInfo_Equals = Equals(Value)
End Function

Private Property Get FileSystemInfo_Exists() As Boolean
    FileSystemInfo_Exists = Exists
End Property

Private Property Get FileSystemInfo_Extension() As String
    FileSystemInfo_Extension = Extension
End Property

Private Property Get FileSystemInfo_FullName() As String
    FileSystemInfo_FullName = FullName
End Property

Private Function FileSystemInfo_GetHashCode() As Long
    FileSystemInfo_GetHashCode = GetHashCode
End Function

Private Property Set FileSystemInfo_LastAccessTime(ByVal RHS As Variant)
    Set LastAccessTime = RHS
End Property

Private Property Get FileSystemInfo_LastAccessTime() As Variant
    Set FileSystemInfo_LastAccessTime = LastAccessTime
End Property

Private Property Let FileSystemInfo_LastAccessTime(ByVal RHS As Variant)
    LastAccessTime = RHS
End Property

Private Property Set FileSystemInfo_LastAccessTimeUtc(ByVal RHS As Variant)
    Set LastAccessTimeUtc = RHS
End Property

Private Property Let FileSystemInfo_LastAccessTimeUtc(ByVal RHS As Variant)
    LastAccessTimeUtc = RHS
End Property

Private Property Get FileSystemInfo_LastAccessTimeUtc() As Variant
    Set FileSystemInfo_LastAccessTimeUtc = LastAccessTimeUtc
End Property

Private Property Set FileSystemInfo_LastWriteTime(ByVal RHS As Variant)
    Set LastWriteTime = RHS
End Property

Private Property Let FileSystemInfo_LastWriteTime(ByVal RHS As Variant)
    LastWriteTime = RHS
End Property

Private Property Get FileSystemInfo_LastWriteTime() As Variant
    Set FileSystemInfo_LastWriteTime = LastWriteTime
End Property

Private Property Set FileSystemInfo_LastWriteTimeUtc(ByVal RHS As Variant)
    Set LastWriteTimeUtc = RHS
End Property

Private Property Let FileSystemInfo_LastWriteTimeUtc(ByVal RHS As Variant)
    LastWriteTimeUtc = RHS
End Property

Private Property Get FileSystemInfo_LastWriteTimeUtc() As Variant
    Set FileSystemInfo_LastWriteTimeUtc = LastWriteTimeUtc
End Property

Private Property Get FileSystemInfo_Name() As String
    FileSystemInfo_Name = Name
End Property

Private Sub FileSystemInfo_Refresh()
    Call Refresh
End Sub

Private Function FileSystemInfo_ToString() As String
    FileSystemInfo_ToString = ToString
End Function

